// Author : Pierre Vuillemin (2016)
// License : CECILL

function out = sopi_parse(var)
   // Main interface for parsing a sopiVar and determining its properties (i.e. 
   // convexity, concavity, etc.)
   // 
   // Input
   // -----
   // var (sopiVar): the variable to parse
   // 
   // Returns
   // -------
   // out (list): a m x n list which entries contain an object representing the
   // properties of the entry.
   //
   // Example
   // -------
   // sopi_begin()
   // x = sopi_var(3);
   // out = sopi_parse(x);
   // out(1)(1).sets // sets to which the first entry of x belongs
   // x.isLinear // internally calls sopi_parse 
   //
   // Notes
   // -----
   // The parsing in only sufficient as some properties might be too difficult
   // to determine for some functions.
   SETS     = "sets"
   CONVEX   = "convex"
   CONCAVE  = "concave"
   DIFF     = "diff"
   POLY     = "poly"
   RAT      = "rat"
   DATA     = "data"
   out      = sopi_recursiveParse(var)
endfunction

// sopi_recursiveParse .........................................................
// parses a sopiVar to determine its properties, such as convexity, concavity, etc.
function out = sopi_recursiveParse(var)
   if var.isElementary then
      belongsTo   = [POLY;CONVEX;CONCAVE;RAT;DIFF]
      data        = struct("order",1,"dorder",0)
      out         = sopi_newMatProp(size(var), SETS, belongsTo, DATA, data)
   else
      select var.operator
      case "constant"
         belongsTo   = [POLY;CONVEX;CONCAVE;RAT;DIFF]
         data        = struct("order",0,"dorder",0)
         si          = sign(var.args(1))
         [m, n]      = size(var)
         out         = sopi_newMatProp([m,n], SETS, belongsTo, DATA, data)
         for i = 1:m
            for j = 1:n
               out(i)(j).data.sign = si(i,j)
            end
         end         
      case "sum"
         p1    = sopi_parse(var.args(1))
         p2    = sopi_parse(var.args(2))
         out   = sopi_addProp(p1,p2)
      case "times"
         p1    = sopi_parse(var.args(1))
         p2    = sopi_parse(var.args(2))
         out   = sopi_mulProp(p1,p2)
      else
         error("not handled yet")
      end
   end
endfunction

// sopi_mulProp ................................................................
// returns a matrix of sets of properties of suitable size corresponding to the
// properties of the sopiVar obtained after matrix multiplication of two sopiVar.
function P = sopi_mulProp(P1,P2)
   m = size(P1)
   n = size(P1(1))
   o = size(P2(1))
   P = sopi_newMatProp([m,o])
   for i = 1:m
      for j = 1:o
         for k = 1:n
            newP = sopi_multiplicationRules(P1(i)(k),P2(k)(j))
            if k~=1
               P(i)(j) = sopi_additionRules(P(i)(j), newP)
            else
               P(i)(j) = newP
            end
         end
      end
   end
endfunction

// sopi_multiplicationRules ....................................................
// returns a set of properties obtained after the multiplication of two variables
// with some given properties. For instance, a negative constant times a convex 
// function is concave, etc.
// Obviously, those are note exhaustives.
function p = sopi_multiplicationRules(p1,p2)
   p = sopi_newProp()
   if p1.isConvex & p2.isConstant then
      if p2.sign > 0 
         p.sets = [p.sets;CONVEX]
      else
         p.sets = [p.sets;CONCAVE]
      end
   elseif p2.isConvex & p1.isConstant then
      if p1.sign > 0 
         p.sets = [p.sets;CONVEX]
      else
         p.sets = [p.sets;CONCAVE]
      end
   elseif p1.isConcave & p2.isConstant then
      if p2.sign > 0 then
         p.sets = [p.sets; CONCAVE]
      else
         p.sets = [p.sets;CONVEX]
      end
   elseif p2.isConvace & p1.isConstant then
      if p1.sign > 0 then
         p.sets = [p.sets;CONCAVE]
      else
         p.sets = [p.sets;CONVEX]
      end
   end
   if p1.isPoly & p2.isPoly then
      p.sets         = [p.sets;POLY]
      p.data.order   = p1.polyOrder + p2.polyOrder;
      if p.polyOrder <= 1
         p.sets = [p.sets;CONVEX;CONCAVE]
      end
   end
   if p1.isRat & p2.isRat then
      p.sets         = [p.sets;RAT]
      p.data.dorder  = p1.data.dorder + p2.data.dorder
   end
   if p1.isDiff & p2.isDiff then
      p.sets         = [p.sets;DIFF]
   end
endfunction

// sopi_addProp ................................................................
// applies, element-wise, the additions rules of two sets of properties.
function P = sopi_addProp(P1,P2)
   m = size(P1)
   n = size(P1(1))
   P = sopi_newMatProp([m,n])
   for i = 1:m
      for j = 1:n
         P(i)(j) = sopi_additionRules(P1(i)(j),P2(i)(j))
      end
   end
endfunction

// sopi_additionRules ..........................................................
// returns a set of properties obtained after the addition of two sets of 
// properties corresponding to some sopiVar. For instance, the sum of two
// convex functions is also convex, etc.
function p = sopi_additionRules(p1, p2)
   p = sopi_newProp()
   if p1.isConvex & p2.isConvex then
      p.sets         = [p.sets;CONVEX]
   end
   if p1.isConcave & p2.isConcave then
      p.sets         = [p.sets;CONCAVE]
   end
   if p1.isPoly & p2.isPoly then
      p.sets         = [p.sets;POLY]
      p.data.order   = max(p1.polyOrder,p2.polyOrder);
   end
   if p1.isRat & p2.isRat then
      p.sets         = [p.sets;RAT]
      p.data.dorder  = max(p1.data.dorder,p2.data.dorder)
   end
   if p1.isDiff & p2.isDiff then
      p.sets         = [p.sets;DIFF]
   end
endfunction

// sopi_newMatProp .............................................................
// creates a matrix of sopiProp that gathers the properties of each entries in 
// a matrix sopiVar.
function P = sopi_newMatProp(s, varargin)
   P = list()
   for i= 1:s(1)
      if size(P) < s(1)
         P($+1) = list()
      end
      for j = 1:s(2)
         if size(P(i)) < s(2)
            P(i)($+1)   = sopi_newProp(varargin(:))
         else
            P(i)(j)     = sopi_newProp(varargin(:))
         end
      end
   end
endfunction

// sopi_newProp ................................................................
// creates a new object sopiProp containing a set of properties for a scalar 
// variable. Properties can be convexity, concavity, etc.
function p = sopi_newProp(varargin)
   objType  = "sopiProp"
   objAtt   = list("sets",[],"data",struct())
   p        = sopi_makeObject(objType, objAtt)
   p        = sopi_fillList(p, varargin)
endfunction
